# Wrapping a third party package

The floats package is an example of using a third party package without any local Go code. Only two function from the third party package are wrapped by rgo to demonstrate the use of the `AllowedFunc` attribute in the rgo configuration.

The package is generated by defining a `go.mod` file with
```
$ go mod init floats
```
running `rgo init` with an argument pointing to the [Gonum floats package](https://pkg.go.dev/gonum.org/v1/gonum/floats?tab=doc),
```
$ rgo init gonum.org/v1/gonum/floats
```
and then editing the `rgo.json` file to regex-match the desired functions in the Go floats package, `floats.CumProd` and `floats.CumSum`.
```
{
	"PkgPath": "gonum.org/v1/gonum/floats",
	"AllowedFuncs": "^Cum(Sum|Prod)$",
	"Words": null,
	"LicenseDir": "LICENSE",
	"LicensePattern": "(LICEN[SC]E|COPYING)(\\.(txt|md))?$"
}
```

Once this is done, the wrapper code can be generated by running the build subcommand.
```
$ rgo build
```
This will generate the Go, C and R wrapper code for the R package, and collate all the licenses in the source package into the `LicenseDir` directory. At this stage the `DESCRIPTION` file should be edited and non-relevant licenses should be removed.

The package can now be installed.
```
$ R CMD INSTALL .
```

The Gonum floats package was not written with R use in mind, but it can demonstrate some of the benefits and dangers of using Go code to manipulate R vectors. The `floats.CumProd` and `floats.CumSum` write their results to preallocated memory for efficiency, so the obvious invocation of the functions results in mutation of the parameter.
```
> library(floats)
> a <- c(1,2,3,4)
> floats::cum_sum(a, a)
[1]  1  3  6 10
> a
[1]  1  3  6 10
> floats::cum_prod(a, a)
[1]   1   3  18 180
> a
[1]   1   3  18 180
```

This may be useful to reduce allocations, though the benefit is lost with these rgo-wrapped function because rgo-wrapped functions that return slices must allocate a copy to satisfy Go runtime rules dealing with handing pointers to C code (it is not allowed for C to retain Go allocations). For a more useful example of in-place mutation see the sort example.

The Gonum floats package requires that parameter length match, otherwise the functions panic. Panics are recovered by the rgo wrapper and converted to an R error.
```
> dst <- vector(mode="double", length=1)
> floats::cum_prod(dst, a)
Error in floats::cum_prod(dst, a) : 
  floats: destination slice length does not match input
```
