# Wrapping a standard library package

The sort package is an example of using a Go standard package without any local Go code. Only three functions from the standard library package are wrapped by rgo to demonstrate the use of the `AllowedFunc` attribute in the rgo configuration, and the utility of the known word attribute is also demonstrated.

The package is generated by defining a `go.mod` file with
```
$ go mod init sort
```
running `rgo init` with an argument pointing to the [Go standard library sort package](https://pkg.go.dev/sort?tab=doc),
```
$ rgo init sort
```
and then editing the `rgo.json` file to regex-match the desired functions in the Go sort package, `sort.Float64s`, `sort.Float64sAreSorted` and `sort.SearchFloat64s`. To help the R code generation to properly break word in Go's camel-case naming covention, we also give "Float64s" as a word (otherwise the camel-case splitter gives back "float_64s").
```
{
	"PkgPath": "sort",
	"AllowedFuncs": "(Float64s|Strings)",
	"Words": ["Float64s"],
	"LicenseDir": "LICENSE",
	"LicensePattern": "(LICEN[SC]E|COPYING)(\\.(txt|md))?$"
}
```

Once this is done, the wrapper code can be generated by running the build subcommand.
```
$ rgo build
```
This will generate the Go, C and R wrapper code for the R package, and collate all the licenses in the source package into the `LicenseDir` directory. At this stage the `DESCRIPTION` file should be edited and non-relevant licenses should be removed.

The package can now be installed.
```
$ R CMD INSTALL .
```

With the package installed, we can now sort vectors of doubles in place, confirm they are sorted and perform binary search for elements. It should be noted that Go indexing is zero-based while R indexing is one-based. This means that indexes must be adjusted from function that are returning native index values, such as `search_float64s`.

```
> library(sort)
> a <- runif(10)
> a
 [1] 0.10797438 0.18757952 0.72708865 0.73772311 0.63402962 0.02649586
 [7] 0.26504999 0.66569101 0.10191473 0.71515159
> sort::float64s_are_sorted(a)
[1] FALSE
> sort::float64s(a)
NULL
> a
 [1] 0.02649586 0.10191473 0.10797438 0.18757952 0.26504999 0.63402962
 [7] 0.66569101 0.71515159 0.72708865 0.73772311
> sort::float64s_are_sorted(a)
[1] TRUE
> sort::search_float64s(a, 0.25)
[1] 4
> a[sort::search_float64s(a, 0.25)+1]
[1] 0.26505
```

Note that the sort-in-place behaviour does not work for the string functions. This is because R `character` vectors are not stored in memory in the same layout as Go `[]string` values. In fact the in-place bahaviour only works for `[]int32`/`[]rune` (`integer`), `[]float64` (`double`), `[]uint8`/`[]byte` (`raw`) and `[]complex128` (`complex`). All other types must be copied in the call into Go due to layout differences and so mutations in the Go code only affect the copy, not the original R vector.

```
> a <- c("one", "two", "three")
> sort::strings_are_sorted(a)
[1] FALSE
> sort::strings(a)
NULL
> sort::strings_are_sorted(a)
[1] FALSE
> a
[1] "one"   "two"   "three"
> a <- base::sort(a)
> a
[1] "one"   "three" "two"  
> sort::strings_are_sorted(a)
[1] TRUE
> sort::search_strings(a, "seven")
[1] 1
> a[sort::search_strings(a, "seven")+1]
[1] "three"
```
